
// Web Interface Plug-In v1.1
// (c) 2021 CrownSoft
// License: GPLv2

extern "C"
{
#include "mongoose.h"
}

#include "../../include/ARGeneralPluginHelper.h"
#include "resource.h"
#include <process.h>
#include "ARRemote.h"

#include <string>
#include <codecvt>

char* indexHtmContent = NULL;
char* jsContent = NULL;
char* cssContent = NULL;

struct mg_mgr mgr;

// {DAF62A89-56E3-4D58-9E1E-7D39D1FF0DEA}
RGUID pluginID = { 0xdaf62a89, 0x56e3, 0x4d58, { 0x9e, 0x1e, 0x7d, 0x39, 0xd1, 0xff, 0xd, 0xea } };

struct PluginState
{
	int posX;
	int posY;
	bool visible;

	// custom members here...
	wchar_t userName[128];
	wchar_t password[128];
	wchar_t port[16];
};

// detect if struct padding has been changed. (change of compiler)
// if changed, the plugin will not accept older configurations from user.
constexpr size_t sizeOfPluginState = sizeof(PluginState);
static_assert(sizeof(PluginState) == 556, "PluginState structure size is not correct");

class WebInterfacePlugin : public ARGeneralPluginHelper<PluginState>, public TButtonListener
{
protected:
	HANDLE hThread;
	volatile bool threadShouldStop;

	TLabel* lblUserName;
	TLabel* lblPassword;
	TLabel* lblPort;
	TLabel* lblVer;
	TEditBox* editUserName;
	TPasswordBox* editPassword;
	TNumericField* editPort;
	TButton* btnUpdate;

	std::string currentUserName;
	std::string currentPassword;
	std::string currentServerUrl;

	std::string buildJsonContent()
	{
		std::string jsonContent = "{\"inputDevices\": [";

		//C++11 format converter
		std::wstring_convert<std::codecvt_utf8_utf16<wchar_t>> convert;

		const int inputDeviceCount = systemAPI->GetInputDeviceCount();

		int selectedInputDeviceIndex = 0;
		SRSTR selectedInputDeviceName;
		systemAPI->GetCurrentInputDeviceName(&selectedInputDeviceName);

		for (int i = 0; i < inputDeviceCount; i++)
		{
			SRSTR deviceName;
			systemAPI->GetInputDeviceNameByIndex(i, &deviceName);

			if (wcscmp(selectedInputDeviceName, deviceName) == 0)
				selectedInputDeviceIndex = i;

			std::string item("\"");
			item += convert.to_bytes(deviceName); // convert to ansi and append.
			item += "\"";

			if((i+1) != inputDeviceCount)
				item += ",";

			jsonContent += item;
		}

		jsonContent += "] ,\"selectedInputDevice\" : ";
		jsonContent += std::to_string(selectedInputDeviceIndex);
		jsonContent += ",\"outputDevices\": [";

		const int outputDeviceCount = systemAPI->GetOutputDeviceCount();

		int selectedOutputDeviceIndex = 0;
		SRSTR selectedOutputDeviceName;
		systemAPI->GetCurrentOutputDeviceName(&selectedOutputDeviceName);

		for (int i = 0; i < outputDeviceCount; i++)
		{
			SRSTR deviceName;
			systemAPI->GetOutputDeviceNameByIndex(i, &deviceName);

			if (wcscmp(selectedOutputDeviceName, deviceName) == 0)
				selectedOutputDeviceIndex = i;

			std::string item("\"");
			item += convert.to_bytes(deviceName); // convert to ansi and append.
			item += "\"";

			if ((i + 1) != outputDeviceCount)
				item += ",";

			jsonContent += item;
		}

		jsonContent += "] ,\"selectedOutputDevice\" : ";
		jsonContent += std::to_string(selectedOutputDeviceIndex);

		jsonContent += ",\"processAudio\" : ";
		
		COPYDATASTRUCT cds;
		cds.dwData = ARIPCMessages::IS_PROCESSING_ENABLED;
		cds.cbData = 0;
		cds.lpData = NULL;

		HWND mainWindow;
		systemAPI->GetApplicationWindow(&mainWindow);

		const bool processingEnabled = (::SendMessageW(mainWindow, WM_COPYDATA, (WPARAM)0, (LPARAM)&cds) == 1) ? true : false;
		jsonContent += processingEnabled ? "true," : "false,";
		
		jsonContent += "\"streamGain\" : ";

		cds.dwData = ARIPCMessages::GET_STREAM_GAIN;
		cds.cbData = 0;
		cds.lpData = NULL;
		jsonContent += std::to_string((int)::SendMessageW(mainWindow, WM_COPYDATA, (WPARAM)0, (LPARAM)&cds));

		jsonContent += ",\"isStreaming\" : ";
		jsonContent += (systemAPI->IsStreaming() == 0) ? "false":"true";
		jsonContent += "}";

		return jsonContent;
	}

	void processParameterChange(int paramIndex, int paramValue)
	{
		if (paramIndex == 1) // input device
		{
			SRSTR deviceID;
			systemAPI->GetInputDeviceIDByIndex(paramValue, &deviceID);
			systemAPI->SetCurrentInputDevice(deviceID);
		}
		else if (paramIndex == 2) // output device
		{
			SRSTR deviceID;
			systemAPI->GetOutputDeviceIDByIndex(paramValue, &deviceID);
			systemAPI->SetCurrentOutputDevice(deviceID);
		}
		else if (paramIndex == 3) // process audio
		{
			int value = paramValue;

			COPYDATASTRUCT cds;
			cds.dwData = ARIPCMessages::SET_PROCESSING_ENABLE;
			cds.cbData = sizeof(int);
			cds.lpData = &value;

			HWND mainWindow;
			systemAPI->GetApplicationWindow(&mainWindow);

			::SendMessageW(mainWindow, WM_COPYDATA, (WPARAM)0, (LPARAM)&cds);
		}
		else if (paramIndex == 4) // Stream gain
		{
			int streamGain = paramValue;

			if (streamGain > 100)
				streamGain = 100;
			else if (streamGain < 0)
				streamGain = 0;

			COPYDATASTRUCT cds;
			cds.dwData = ARIPCMessages::SET_STREAM_GAIN;
			cds.cbData = sizeof(int);
			cds.lpData = &streamGain;

			HWND mainWindow;
			systemAPI->GetApplicationWindow(&mainWindow);

			::SendMessageW(mainWindow, WM_COPYDATA, (WPARAM)0, (LPARAM)&cds);
		}
		else if (paramIndex == 5) // Start/Stop
		{
			COPYDATASTRUCT cds;
			cds.dwData = (paramValue == 0) ? ARIPCMessages::REPEATER_STOP : ARIPCMessages::REPEATER_START;
			cds.cbData = 0;
			cds.lpData = NULL;

			HWND mainWindow;
			systemAPI->GetApplicationWindow(&mainWindow);

			::SendMessageW(mainWindow, WM_COPYDATA, (WPARAM)0, (LPARAM)&cds);
		}
	}

	void httpEventHandler(struct mg_connection* c, int ev, void* ev_data)
	{
		if (ev == MG_EV_HTTP_MSG)
		{
			struct mg_http_message* hm = (struct mg_http_message*)ev_data;

			char user[256], pass[256];
			mg_http_creds(hm, user, 256, pass, 256);

			if (user[0] != '\0' && pass[0] != '\0')
			{
				if (strcmp(user, currentUserName.c_str()) == 0 && strcmp(pass, currentPassword.c_str()) == 0)
				{
					if (mg_http_match_uri(hm, "/"))
					{
						mg_http_reply(c, 200, "Cache-Control: no-cache\r\nContent-Type: text/html\r\n", indexHtmContent);
					}
					else if (mg_http_match_uri(hm, "/main.js"))
					{
						mg_http_reply(c, 200, "Cache-Control: no-cache\r\nContent-Type: text/javascript\r\n", jsContent);
					}
					else if (mg_http_match_uri(hm, "/style.css"))
					{
						mg_http_reply(c, 200, "Cache-Control: no-cache\r\nContent-Type: text/css\r\n", cssContent);
					}
					else if (mg_http_match_uri(hm, "/state.json"))
					{
						std::string jsonContent = this->buildJsonContent();
						mg_http_reply(c, 200, "Cache-Control: no-cache\r\nContent-Type: application/json\r\n", jsonContent.c_str());
					}
					else if (mg_http_match_uri(hm, "/update"))
					{
						const int size = 16;
						char buffer[size];
						if (mg_http_get_var(&hm->body, "paramIndex", buffer, size) >= 1)
						{
							const int paramIndex = atoi(buffer);
							if (mg_http_get_var(&hm->body, "paramVal", buffer, size) >= 1)
							{
								const int paramValue = atoi(buffer);
								//printf("param index:%d, param value:%d\n", paramIndex, paramValue);

								this->processParameterChange(paramIndex, paramValue);
								mg_http_reply(c, 200, "Cache-Control: no-cache\r\nContent-Type: text/plain\r\n", "OK");
								return;
							}						
						}
						
						mg_http_reply(c, 200, "Cache-Control: no-cache\r\nContent-Type: text/plain\r\n", "ERROR");
						return;
					}
					else
					{
						mg_printf(c, "%s", "HTTP/1.1 404 Not Found\r\nContent-Length: 0\r\n\r\n");
					}

					return;
				}
			}

			mg_printf(c, "%s", "HTTP/1.1 401 Denied\r\nWWW-Authenticate: Basic\r\nContent-Length: 0\r\n\r\n");
		}
	}

	static void httpEventHandlerCallback(struct mg_connection* c, int ev, void* ev_data, void* fn_data)
	{
		WebInterfacePlugin* plugin = (WebInterfacePlugin*)fn_data;
		plugin->httpEventHandler(c, ev, ev_data);
	}

	// assume file is ansi
	void LoadFileFromResourceA(HMODULE handle, int resourceID, DWORD* size, const char** data)
	{
		HRSRC rc = ::FindResourceW(handle, MAKEINTRESOURCEW(resourceID), RT_HTML);
		HGLOBAL rcData = ::LoadResource(handle, rc);
		*size = ::SizeofResource(handle, rc);
		*data = (const char*)::LockResource(rcData);
	}

	// assume file is ansi.
	// caller must free the returned pointer.
	char* ReadResourceFileA(HMODULE handle, int resourceID)
	{
		DWORD size = 0;
		const char* data = NULL;
		LoadFileFromResourceA(handle, resourceID, &size, &data);

		// The text stored in the resource might not be NULL terminated.
		char* buffer = (char*)::malloc(size + 1);
		::memcpy(buffer, data, size);
		buffer[size] = 0; // NULL terminator

		return buffer;
	}

public:
	int OnInitialize(int hostID, void* hostParam) override
	{
		// assign default values to custom members of the plugin state
		::wcscpy(pluginState.userName, L"admin");
		::wcscpy(pluginState.password, L"1234");
		::wcscpy(pluginState.port, L"8000");

		std::wstring_convert<std::codecvt_utf8_utf16<wchar_t>> convert;

		currentUserName = convert.to_bytes(pluginState.userName);
		currentPassword = convert.to_bytes(pluginState.password);
		currentServerUrl = "http://0.0.0.0:" + convert.to_bytes(pluginState.port);

		indexHtmContent = ReadResourceFileA(hInstance, IDR_HTML1);
		jsContent = ReadResourceFileA(hInstance, IDR_HTML2);
		cssContent = ReadResourceFileA(hInstance, IDR_HTML3);
	
		return 1;
	}

	void OnFree() override
	{
		::free(indexHtmContent);
		::free(jsContent);
		::free(cssContent);
	}

	void OnCreateUI(ARServiceFactory* stdServiceFactory, ARGUIService* guiService) override
	{
		pluginWindow->SetSize(250, 190);

		guiService->CreateLabel(pluginWindow, 0, &lblUserName);
		lblUserName->SetPosition(20, 28);
		lblUserName->SetText(SRSTR(L"User Name"));

		guiService->CreateLabel(pluginWindow, 0, &lblPassword);
		lblPassword->SetPosition(20, 68);
		lblPassword->SetText(SRSTR(L"Password"));

		guiService->CreateLabel(pluginWindow, 0, &lblPort);
		lblPort->SetPosition(20, 108);
		lblPort->SetText(SRSTR(L"Server Port"));

		guiService->CreateEditBox(pluginWindow, &editUserName);
		editUserName->SetSize(120, 0);
		editUserName->SetPosition(111, 28);
		
		guiService->CreatePasswordBox(pluginWindow, &editPassword);
		editPassword->SetSize(120, 0);
		editPassword->SetPosition(111, 67);

		guiService->CreateNumericField(pluginWindow, &editPort);
		editPort->SetSize(120, 0);
		editPort->SetPosition(111, 106);

		guiService->CreateButton(pluginWindow, &btnUpdate);
		btnUpdate->SetPosition(161, 146);
		btnUpdate->SetSize(70, btnUpdate->GetHeight());
		btnUpdate->SetText(SRSTR(L"Update"));
		btnUpdate->SetListener(this);

		guiService->CreateLabel(pluginWindow, 0, &lblVer);
		lblVer->SetPosition(20, 151);
		lblVer->SetText(SRSTR(L"Web Interface v1.1"));
		lblVer->SetTextColor(RGB(100, 100, 100));
	}

	void OnDestroyUI() override
	{
		lblUserName->Release();
		lblPassword->Release();
		lblPort->Release();
		editUserName->Release();
		editPassword->Release();
		editPort->Release();
		btnUpdate->Release();
		lblVer->Release();
	}

	void OnStateLoad() override
	{
		std::wstring_convert<std::codecvt_utf8_utf16<wchar_t>> convert;

		currentUserName = convert.to_bytes(pluginState.userName);
		currentPassword = convert.to_bytes(pluginState.password);
		currentServerUrl = "http://0.0.0.0:" + convert.to_bytes(pluginState.port);
	}

	void OnApplyPluginStateToUI() override
	{
		SRSTR userName(pluginState.userName);
		editUserName->SetText(userName);

		SRSTR password(pluginState.password);
		editPassword->SetText(password);

		SRSTR port(pluginState.port);
		editPort->SetText(port);
	}

	void setEnableComponents(bool enable)
	{
		btnUpdate->SetEnable(enable);
		editUserName->SetEnable(enable);
		editPassword->SetEnable(enable);
		editPort->SetEnable(enable);
	}

	void STDCALL OnButtonPress(TButton* button) override
	{
		if (button == btnUpdate)
		{
			this->setEnableComponents(false);

			SRSTR userName;
			editUserName->GetText(&userName);

			if (userName->GetLength() < 3)
			{
				systemAPI->ShowMessage(ARMsgIconTypes::MSG_ERROR, SRSTR(L"Error"), SRSTR(L"Invalid user name entered."), ARMsgButtonTypes::MSG_BUTTON_OK);
				this->setEnableComponents(true);
				return;
			}

			SRSTR password;
			editPassword->GetText(&password);

			if (password->GetLength() < 3)
			{
				systemAPI->ShowMessage(ARMsgIconTypes::MSG_ERROR, SRSTR(L"Error"), SRSTR(L"Invalid password entered."), ARMsgButtonTypes::MSG_BUTTON_OK);
				this->setEnableComponents(true);
				return;
			}

			SRSTR port;
			editPort->GetText(&port);

			if(port->GetLength() < 1)
			{
				systemAPI->ShowMessage(ARMsgIconTypes::MSG_ERROR, SRSTR(L"Error"), SRSTR(L"Invalid port value entered."), ARMsgButtonTypes::MSG_BUTTON_OK);
				this->setEnableComponents(true);
				return;
			}

			//C++11 format converter
			std::wstring_convert<std::codecvt_utf8_utf16<wchar_t>> convert;
			
			std::string strPort = convert.to_bytes(port);
			const int intPort = atoi(strPort.c_str());

			if ((intPort < 1) || (intPort > 65535))
			{
				systemAPI->ShowMessage(ARMsgIconTypes::MSG_ERROR, SRSTR(L"Error"), SRSTR(L"Invalid port value entered."), ARMsgButtonTypes::MSG_BUTTON_OK);
				this->setEnableComponents(true);
				return;
			}

			this->stopWebServer();

			::wcscpy(pluginState.userName, userName);
			::wcscpy(pluginState.password, password);
			::wcscpy(pluginState.port, port);

			currentUserName = convert.to_bytes(userName);
			currentPassword = convert.to_bytes(password);
			currentServerUrl = "http://0.0.0.0:" + convert.to_bytes(port);

			this->startWebServer();

			this->setEnableComponents(true);
		}
	}

	void startWebServer()
	{
		threadShouldStop = false;
		hThread = (HANDLE)::_beginthreadex(NULL, 0, WebInterfacePlugin::Thread_Proc, this, CREATE_SUSPENDED, NULL); // create in suspended state so we can assign to handle variable.
		::ResumeThread(hThread);
	}

	void stopWebServer()
	{
		threadShouldStop = true;
		::WaitForSingleObject(hThread, INFINITE);
	}

	static unsigned __stdcall Thread_Proc(void* lpParameter)
	{
		if (lpParameter == 0) // for safe!
			return 0;

		WebInterfacePlugin* plugin = (WebInterfacePlugin*)lpParameter;

		mg_mgr_init(&mgr); // Initialise event manager
		mg_http_listen(&mgr, plugin->currentServerUrl.c_str(), httpEventHandlerCallback, plugin); // Create HTTP listener
	
		while (!plugin->threadShouldStop)
		{
			mg_mgr_poll(&mgr, 100);
		}

		mg_mgr_free(&mgr);
		
		return 0;
	}

	void OnEnable() override
	{
		startWebServer();
	}

	void OnDisable() override
	{
		stopWebServer();
	}

	void STDCALL GetID(RGUID** id) override
	{
		*id = &pluginID;
	}

	const wchar_t* GetPluginName() override
	{
		return L"Web Interface";
	}

	const wchar_t* GetPluginAuthor() override
	{
		return L"CrownSoft";
	}

	const wchar_t* GetPluginWindowText() override
	{
		return L"Web Interface Config";
	}

	const wchar_t* GetPluginMenuText() override
	{
		return L"Web Interface Plug-In";
	}

	IMPLEMENTS_RUNKNOWN_METHODS
};

EXPORT_PLUGIN(WebInterfacePlugin)
SELF_REGISTER_PLUGIN("{DAF62A89-56E3-4D58-9E1E-7D39D1FF0DEA}")